#import "RadonView.h"

@implementation RadonView

- initWithFrame:(NSRect)frameRect
{
    int i;

    [super initWithFrame:frameRect];
    bitmap = nil;
    nx = 0;
    ny = 0;
    maxz = 0;
    
    imageData = NULL;
    filterData = NULL;
    transformData = NULL;

    openPanel = nil;
    savePanel = nil;

    /* Init colors */
    for(i=0; i<64; i++){
	heatred[i] = 0.0;
	heatgreen[i] = 0.0;
	heatblue[i] = i*4;
    }

    for(i=0; i<64; i++) {
	heatred[64+i] = i*4;
	heatgreen[64+i] = 0;
	heatblue[64+i] = 255 - i*4;
    }
    
    for(i=0; i<64; i++) {
	heatred[128+i] = 255;
	heatgreen[128+i] = i*4;
	heatblue[128+i] = 0;
    }
    
    for(i=0; i<64; i++) {
	heatred[192+i] = 255;
	heatgreen[192+i] = 255;
	heatblue[192+i] = i*4;
    }
    
    return self;
}

- (void)dealloc
{
    if(bitmap) [bitmap release];
    if(imageData) im_free2D((char **)imageData);
    if(transformData) im_free2D((char **)transformData);
    return;
}


- openImage:sender
{
    NSString *fileName;

    if(!openPanel) openPanel = [NSOpenPanel new];
    if([openPanel runModal] == NSCancelButton) return nil;

    fileName = [[openPanel filenames] objectAtIndex:0];

    [self openFile:fileName];

    return self;
}

    

- openFile:(NSString *)fileName
{
    FILE *imagefd;
    int i,ndata;
    unsigned char *newdata[1];
    char magic[2];
    float *floatptr;

    /* Open and read fileName */
    if(!(imagefd=fopen([fileName cString],"r"))) {
	fprintf(stderr,"Can't open image file: %s\n",[fileName cString]);
	exit(1);
    }

    fscanf(imagefd,"%c%c",&magic[0],&magic[1]);
    if(magic[0]!='P') {
	fprintf(stderr,"Input %s is not a PGM file\n",[fileName cString]);
	exit(1);
    }

    if((magic[1]!='2') && (magic[1]!='5')) {
	fprintf(stderr,"Input %s is not a PGM file\n",[fileName cString]);
	exit(1);
    }

    fscanf(imagefd,"%d %d %d",&nx,&ny,&maxz);
    ndata = nx*ny;
    
    if(imageData) im_free2D((char **)imageData);
    imageData = (float **)im_alloc2D(nx,ny);
    floatptr = imageData[0];

    for(i=0; i<ndata; i++) {
	fscanf(imagefd,"%f",floatptr+i);
    }

    fclose(imagefd);

    /* Allocate bitmap data */
    newdata[0] = (unsigned char *)malloc(ndata*sizeof(unsigned char));

    for(i=0; i<ndata; i++) {
	newdata[0][i] = (unsigned char)(floatptr[i]*255.0/maxz);
    }

    /* Make a new bitmap */
    if(bitmap) [bitmap release];
    bitmap = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:newdata
				       pixelsWide:nx
				       pixelsHigh:ny
				       bitsPerSample:8
				       samplesPerPixel:1
				       hasAlpha:NO
				       isPlanar:YES
				       colorSpaceName:NSCalibratedWhiteColorSpace
				       bytesPerRow:nx
				       bitsPerPixel:8];

    if(!bitmap) {
	fprintf(stderr,"Can't make new bitmap from rawdata\n");
	exit(1);
    }

    [self setNeedsDisplay:YES];

    return self;
}


- project:sender
{
    int x, y, a, b, Xcenter, Ycenter;
    int i,ndata,NumBuckets,NumProject;
    float val, Position, *SinTab, *CosTab;
    NSBitmapImageRep *newBitmap;
    float *floatptr,MaxZ,MinZ;
    unsigned char *newdata[1];

    NumBuckets = nx;
    NumProject = ny;
    ndata = nx*ny;

    /* Initialize sin and cos lookup tables */
    SinTab = (float *) malloc((unsigned) sizeof(float) * NumProject);
    CosTab = (float *) malloc((unsigned) sizeof(float) * NumProject);
    for (a = 0; a < NumProject; a++) {
	SinTab[a] = (float) sin((double) (a * M_PI / NumProject));
	CosTab[a] = (float) cos((double) (a * M_PI / NumProject));
    }

    if(transformData) im_free2D((char **)transformData);
    transformData = (float **)im_alloc2D(nx,ny);

    /* Initialize projections to zero */
    for (a = 0; a < NumProject; a++)
	for (b = 0; b < NumBuckets; b++)
	    transformData[a][b] = 0.0;

    /* Project each pixel in the image */
    Xcenter = nx / 2;
    Ycenter = ny / 2;
    for (y = -Ycenter; y < Ycenter; y++) {
	for (x = -Xcenter; x < Xcenter; x++) {
	    val = imageData[y + Ycenter][x + Xcenter];
	    for (a = 0; a < NumProject; a++) {
		Position = x * CosTab[a] + y * SinTab[a];
		/* Calculate coordinates of nearest bucket */
		b = ROUND(Position);
		if (b < 0)
		    b += NumBuckets;
		transformData[a][b] += val;
	    }
	}
    }

    /* Fix 45 and 135 projections  */
    for (b = 0; b < NumBuckets; b++) {
	transformData[NumProject / 4][b] = transformData[1 + NumProject / 4][b];
	transformData[3 * NumProject / 4][b] = transformData[1 + 3 * NumProject / 4][b];
    }


    /* Find MaxZ */
    MaxZ = 0.0;
    MinZ = 1.0e7;
    for(a=0; a<NumProject; a++) {
	for(b=0; b<NumBuckets; b++) {
	    if(transformData[a][b] > MaxZ) MaxZ = transformData[a][b];
	    else if(transformData[a][b] < MinZ) MinZ = transformData[a][b];
	}
    }

    /* Allocate bitmap data */
    newdata[0] = (unsigned char *)malloc(ndata*sizeof(unsigned char));

    floatptr = transformData[0];
    for(i=0; i<ndata; i++) {
	newdata[0][i] = (unsigned char)(255.0*(floatptr[i]-MinZ)/(MaxZ-MinZ));
    }

    newBitmap = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:newdata
				       pixelsWide:nx
				       pixelsHigh:ny
				       bitsPerSample:8
				       samplesPerPixel:1
				       hasAlpha:NO
				       isPlanar:YES
				       colorSpaceName:NSCalibratedWhiteColorSpace
				       bytesPerRow:nx
				       bitsPerPixel:8];

    if(!newBitmap) {
	fprintf(stderr,"Can't make new bitmap from rawdata\n");
	exit(1);
    }

    [viewB setBitmap:newBitmap];

    if([colorizeFlag state])
	[self colorize:heatred :heatgreen :heatblue];

    free(SinTab);
    free(CosTab);
    return self;
}

- (void)filter
{
    int i, a, b,ndata;
    float *filter;
    double *real,*imaginary;
    int Width;
    int NumProject,NumBuckets;

    if(filterData) im_free2D((char **)filterData);
    filterData = (float **)im_alloc2D(nx,ny);

    /* Initialize the filter */
    ndata = nx*ny;
    NumBuckets = nx;
    NumProject = ny;
    Width = nx/2.0;

    filter = (float *) malloc(nx * sizeof(float));

    filter[0] = 0.0;

    for (i = 1; i < Width; i++) {
	filter[nx - i] = filter[i] = (float)i/Width;
    }

    for (i = Width; i <= nx/2; i++) filter[nx-i] = filter[i] = 1.0;

    real = (double *)malloc(nx*sizeof(double));
    imaginary = (double *)malloc(nx*sizeof(double));

    /* Filter each projection */
    for (a = 0; a < ny; a++) {
	for(i=0; i<nx; i++) {
	    real[i] = imageData[a][i];
	    imaginary[i] = 0.0;
	}
	fft(nx,real,imaginary);
	for (b = 0; b < nx; b++) {
	    real[b] *= filter[b];
	    imaginary[b] *= filter[b];
	}
	ifft(nx,real,imaginary);
	for(i=0; i<nx; i++) {
	    filterData[a][i] = real[i];
	}
    }

    free(filter);
    free(real);
    free(imaginary);
}

    


- backProject:sender
{
    int x, y, a, b, Xcenter, Ycenter;
    float val, Position, Scale;
    double *SinTab, *CosTab;
    int NumProject,NumBuckets;
    NSBitmapImageRep *newBitmap;
    float *floatptr,MaxZ,MinZ;
    int i,ndata;
    unsigned char *newdata[1];

    [self filter];

    NumBuckets = nx;
    NumProject = ny;
    ndata = nx*ny;

    /* Initialize sin and cos lookup tables */
    SinTab = (double *) malloc((unsigned) sizeof(double) * NumProject);
    CosTab = (double *) malloc((unsigned) sizeof(double) * NumProject);
    for (a = 0; a < NumProject; a++) {
      SinTab[a] = sin((double) a * M_PI / NumProject);
      CosTab[a] = cos((double) a * M_PI / NumProject);
    }

    if(transformData) im_free2D((char **)transformData);
    transformData = (float **)im_alloc2D(nx,ny);

    /* Initialize output image to zero */
    for (y = 0; y < ny; y++)
	for (x = 0; x < nx; x++)
	    transformData[y][x] = 0;

    /* Project each pixel in the image */
    Scale = (float) (1.0 / (NumProject * NumProject));
    Xcenter = nx / 2;
    Ycenter = ny / 2;
    for (y = -Ycenter; y < Ycenter; y++) {
	for (x = -Xcenter; x < Xcenter; x++) {
	    val = 0.0;
	    for (a = 0; a < NumProject; a++) {
		Position = (float) (x * CosTab[a] + y * SinTab[a]);
		b = ROUND(Position);
		if (b < 0)
		    b += NumBuckets;
		val += filterData[a][b];
	    }
	    transformData[y + Ycenter][x + Xcenter] = val * Scale;
	}
    }

    MaxZ = 0.0;
    MinZ = 1.0e7;
    for(a=0; a<NumProject; a++) {
	for(b=0; b<NumBuckets; b++) {
	    if(transformData[a][b] > MaxZ) MaxZ = transformData[a][b];
	    else if(transformData[a][b] < MinZ) MinZ = transformData[a][b];
	}
    }

    /* Allocate bitmap data */
    newdata[0] = (unsigned char *)malloc(ndata*sizeof(unsigned char));

    floatptr = transformData[0];
    for(i=0; i<ndata; i++) {
	newdata[0][i] = (unsigned char)(255.0*(floatptr[i]-MinZ)/(MaxZ-MinZ));
    }

    newBitmap = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:newdata
				       pixelsWide:nx
				       pixelsHigh:ny
				       bitsPerSample:8
				       samplesPerPixel:1
				       hasAlpha:NO
				       isPlanar:YES
				       colorSpaceName:NSCalibratedWhiteColorSpace
				       bytesPerRow:nx
				       bitsPerPixel:8];

    if(!newBitmap) {
	fprintf(stderr,"Can't make new bitmap from rawdata\n");
	exit(1);
    }

    [viewB setBitmap:newBitmap];

    if([colorizeFlag state])
	[self colorize:heatred :heatgreen :heatblue];

    free(SinTab);
    free(CosTab);
    return self;
}

- saveOutput:sender
{
    NSString *fileName;
    FILE *outfd;
    float MaxZ,MinZ,*floatptr;
    int i,ndata = nx*ny;
    NSData *bitmapTiff;

    if(!savePanel) savePanel = [NSSavePanel new];
    [savePanel setTitle:@"Save"];
    
    if([savePanel runModal] == NSCancelButton) return nil;

    fileName = [savePanel filename];

    /* Save color images as tiffs */
    if([[viewB bitmap] samplesPerPixel] > 1) {
	bitmapTiff = [[viewB bitmap] TIFFRepresentationUsingCompression:NSTIFFCompressionLZW 
				     factor:0];
	[bitmapTiff writeToFile:fileName atomically:NO];
	return self;
    }

    if(!(outfd=fopen([fileName cString],"w"))) {
	fprintf(stderr,"Can't open output file: %s\n",[fileName cString]);
	return self;
    }

    floatptr = transformData[0];

    MaxZ = 0.0;
    MinZ = 1.0e7;
    for(i=0; i<ndata; i++) {
      if(floatptr[i] > MaxZ) MaxZ = floatptr[i];
      if(floatptr[i] < MinZ) MinZ = floatptr[i];
    }

    fprintf(outfd,"P2 %d %d 8192\n",nx,ny);

    for(i=0; i<ndata; i++) {
	fprintf(outfd,"%4d ",(int)(8192.0*(floatptr[i]-MinZ)/(MaxZ-MinZ)));
	if(!((i+1)%16)) fprintf(outfd,"\n");
    }

    /* Close image */
    fclose(outfd);
    
    return self;
}



- colorize:(unsigned char *)redLUT :(unsigned char *)greenLUT :(unsigned char *)blueLUT
{
    int npixels,ncols,nrows,bitsPerSample,samplesPerPixel,isPlanar,bitsPerPixel;
    int ipixel;
    unsigned char *bitmapdata;
    unsigned char *redPlane,*greenPlane,*bluePlane;
    unsigned char *colorPlane[3];
    id colorBitmap;
    id monoBitmap;

    ncols = nx;
    nrows = ny;
    npixels = ncols*nrows;
    monoBitmap = [viewB bitmap];
    bitsPerSample = [monoBitmap bitsPerSample];
    samplesPerPixel = [monoBitmap samplesPerPixel];
    isPlanar = [monoBitmap isPlanar];
    bitsPerPixel = [monoBitmap bitsPerPixel];

    redPlane = (unsigned char *)malloc(nrows*ncols*sizeof(unsigned char));
    greenPlane = (unsigned char *)malloc(nrows*ncols*sizeof(unsigned char));
    bluePlane = (unsigned char *)malloc(nrows*ncols*sizeof(unsigned char));

    bitmapdata = [monoBitmap bitmapData];
    
    for(ipixel=0; ipixel<npixels; ipixel++) {
	redPlane[ipixel] = redLUT[bitmapdata[ipixel]];
	greenPlane[ipixel] = greenLUT[bitmapdata[ipixel]];
	bluePlane[ipixel] = blueLUT[bitmapdata[ipixel]];
    }

    colorPlane[0] = redPlane;
    colorPlane[1] = greenPlane;
    colorPlane[2] = bluePlane;

    colorBitmap = [[NSBitmapImageRep alloc]
		      initWithBitmapDataPlanes:colorPlane
		      pixelsWide:ncols
		      pixelsHigh:nrows
		      bitsPerSample:8
		      samplesPerPixel:3
		      hasAlpha:NO
		      isPlanar:YES
		      colorSpaceName:NSCalibratedRGBColorSpace
		      bytesPerRow:ncols
		      bitsPerPixel:8];

    [viewB setBitmap:colorBitmap];

    return colorBitmap;
}



- (void)setBitmap:(NSBitmapImageRep *)newBitmap
{
    if(newBitmap==nil) return;

    [newBitmap retain];
    if(bitmap) [bitmap release];
    bitmap = newBitmap;
    [self setNeedsDisplay:YES];
    return;
}

- (NSBitmapImageRep *)bitmap
{
    return bitmap;
}


- (void)drawRect:(NSRect)rect
{
    [bitmap drawInRect:[self bounds]];
}


#define GOOD_TRIG
#include "trigtbl.h"

#define SQRT2_2   0.70710678118654752440084436210484
#define SQRT2     1.41421356237309504880168872420968
/* #define SQRT2   2*0.70710678118654752440084436210484 */

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine produces a 1D fourier Transform                    *
/*---------------------------------------------------------------------------*/
void fft(int n,double *real, double *imag) {
    double a,b,c,d;
    double q,r,s,t;
    int i,j,k;
    for (i=1,j=n-1,k=n/2;i<k;i++,j--) {
        a = real[i]; b = real[j];  q=a+b; r=a-b;
        c = imag[i]; d = imag[j];  s=c+d; t=c-d;
        real[i] = (q+t)*.5; real[j] = (q-t)*.5;
        imag[i] = (s-r)*.5; imag[j] = (s+r)*.5;
    }
    fht(real,n);
    fht(imag,n);
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine produces a 1D Inverse Fourier Transform            *
/*---------------------------------------------------------------------------*/
void ifft(int n,double *real, double *imag) {
    double a,b,c,d;
    double q,r,s,t;
    int i,j,k;
    fht(real,n);
    fht(imag,n);
    for (i=1,j=n-1,k=n/2;i<k;i++,j--) {
        a = real[i]; b = real[j];  q=a+b; r=a-b;
        c = imag[i]; d = imag[j];  s=c+d; t=c-d;
        imag[i] = (s+r)*0.5;  imag[j] = (s-r)*0.5;
        real[i] = (q-t)*0.5;  real[j] = (q+t)*0.5;
    }
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine produces a 1D Fast Hartley Transform               *
/*---------------------------------------------------------------------------*/

void fht(REAL *fz,int n) {
    int i,k,k1,k2,k3,k4,kx;
    REAL *fi,*fn,*gi;
    TRIG_VARS;

    for (k1=1,k2=0;k1<n;k1++) {
        REAL a;
        for (k=n>>1; (!((k2^=k)&k)); k>>=1);
        if (k1>k2) {
            a=fz[k1];fz[k1]=fz[k2];fz[k2]=a;
        }
    }
    for ( k=0 ; (1<<k)<n ; k++ );
    k  &= 1;
    if (k==0) {
        for (fi=fz,fn=fz+n;fi<fn;fi+=4) {
            REAL f0,f1,f2,f3;
            f1    = fi[0 ]-fi[1 ];
            f0    = fi[0 ]+fi[1 ];
            f3    = fi[2 ]-fi[3 ];
            f2    = fi[2 ]+fi[3 ];
            fi[2 ] = (f0-f2);
            fi[0 ] = (f0+f2);
            fi[3 ] = (f1-f3);
            fi[1 ] = (f1+f3);
        }
    }
    else {
        for (fi=fz,fn=fz+n,gi=fi+1;fi<fn;fi+=8,gi+=8) {
            REAL s1,c1,s2,c2,s3,c3,s4,c4,g0,f0,f1,g1,f2,g2,f3,g3;
            c1    = fi[0 ] - gi[0 ];
            s1    = fi[0 ] + gi[0 ];
            c2    = fi[2 ] - gi[2 ];
            s2    = fi[2 ] + gi[2 ];
            c3    = fi[4 ] - gi[4 ];
            s3    = fi[4 ] + gi[4 ];
            c4    = fi[6 ] - gi[6 ];
            s4    = fi[6 ] + gi[6 ];
            f1    = (s1 - s2);
            f0    = (s1 + s2);
            g1    = (c1 - c2);
            g0    = (c1 + c2);
            f3    = (s3 - s4);
            f2    = (s3 + s4);
            g3    = SQRT2*c4;
            g2    = SQRT2*c3;
            fi[4 ] = f0 - f2;
            fi[0 ] = f0 + f2;
            fi[6 ] = f1 - f3;
            fi[2 ] = f1 + f3;
            gi[4 ] = g0 - g2;
            gi[0 ] = g0 + g2;
            gi[6 ] = g1 - g3;
            gi[2 ] = g1 + g3;
        }
    }
    if (n<16) return;

    do {
        REAL s1,c1;
        k  += 2;
        k1  = 1  << k;
        k2  = k1 << 1;
        k4  = k2 << 1;
        k3  = k2 + k1;
        kx  = k1 >> 1;
        fi  = fz;
        gi  = fi + kx;
        fn  = fz + n;

        do {
            REAL g0,f0,f1,g1,f2,g2,f3,g3;
            f1     = fi[0 ] - fi[k1];
            f0     = fi[0 ] + fi[k1];
            f3     = fi[k2] - fi[k3];
            f2     = fi[k2] + fi[k3];
            fi[k2]  = f0        - f2;
            fi[0 ]  = f0        + f2;
            fi[k3]  = f1        - f3;
            fi[k1]  = f1        + f3;
            g1     = gi[0 ] - gi[k1];
            g0     = gi[0 ] + gi[k1];
            g3     = SQRT2  * gi[k3];
            g2     = SQRT2  * gi[k2];
            gi[k2]  = g0        - g2;
            gi[0 ]  = g0        + g2;
            gi[k3]  = g1        - g3;
            gi[k1]  = g1        + g3;
            gi    += k4;
            fi    += k4;
        } while (fi<fn);

        TRIG_INIT(k,c1,s1);

        for (i=1;i<kx;i++) {
            REAL c2,s2;
            TRIG_NEXT(k,c1,s1);
            c2 = c1*c1 - s1*s1;
            s2 = 2*(c1*s1);
            fn = fz + n;
            fi = fz +i;
            gi = fz +k1-i;
            do {
                REAL a,b,g0,f0,f1,g1,f2,g2,f3,g3;
                b      = s2*fi[k1] - c2*gi[k1];
                a      = c2*fi[k1] + s2*gi[k1];
                f1     = fi[0 ]    - a;
                f0     = fi[0 ]    + a;
                g1     = gi[0 ]    - b;
                g0     = gi[0 ]    + b;
                b      = s2*fi[k3] - c2*gi[k3];
                a      = c2*fi[k3] + s2*gi[k3];
                f3     = fi[k2]    - a;
                f2     = fi[k2]    + a;
                g3     = gi[k2]    - b;
                g2     = gi[k2]    + b;
                b      = s1*f2    - c1*g3;
                a      = c1*f2    + s1*g3;
                fi[k2] = f0       - a;
                fi[0 ] = f0       + a;
                gi[k3] = g1       - b;
                gi[k1] = g1       + b;
                b      = c1*g2    - s1*f3;
                a      = s1*g2    + c1*f3;
                gi[k2] = g0       - a;
                gi[0 ] = g0       + a;
                fi[k3] = f1       - b;
                fi[k1] = f1       + b;
                gi    += k4;
                fi    += k4;
            } while (fi<fn);
        }
        TRIG_RESET(k,c1,s1);
    } while (k4<n);
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine produces a 1D image access array.                  */
/*---------------------------------------------------------------------------*/
char *im_alloc1D(int nx, int ny) {
   int i, Count, Size;
   char *Data;

   /* Allocate data buffer */
   Count = nx * ny;
   Size = sizeof(float);

   if ((Data = (char *) malloc((unsigned) (Count * Size))) == NULL)
      Error("Could not allocate image buffer");

   for (i = 0; i < Count * Size; i++)
      Data[i] = 0;

   return (Data);
}


/*---------------------------------------------------------------------------*/
/* Purpose:  This routine produces a 2D image access array.                  */
/*---------------------------------------------------------------------------*/
char **im_alloc2D(int nx, int ny) {
    char *Data1, **Data2;
    int y, Count, Size, Offset;

   /* Allocate row access buffer */
   Count = ny;
   Size = sizeof(char *);

   if ((Data2 = (char **) malloc((unsigned) (Count * Size))) == NULL)
      Error("Could not allocate row access buffer");

   /* Initialize row access buffer */
   Offset = nx * sizeof(float);
   Data1 = im_alloc1D(nx,ny);
   for (y = 0; y < Count; y++)
      Data2[y] = Data1 + y * Offset;
   return (Data2);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine frees space from 1D image access array.            */
/*---------------------------------------------------------------------------*/
void im_free1D(char *Data)
{
   free(Data);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine frees space from 2D image access array.            */
/*---------------------------------------------------------------------------*/
void im_free2D(char **Data)
{
   free(Data[0]);
   free(Data);
}

/*---------------------------------------------------------------------------*/
/* Purpose:  This routine prints an error message and aborts.                */
/*---------------------------------------------------------------------------*/
void Error(char *Message)
{
   fprintf(stderr, "%s\n", Message);
   exit(0);
}


@end
